import os
from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from typing import MutableMapping

from .command import Command
from .exc import RunnerError
from .util import abs_path


class Collection(MutableMapping):

    """A collection of commands."""

    def __init__(self, commands):
        self.commands = commands

    @classmethod
    def load_from_module(cls, module):
        module = cls.get_module(module)

        commands = {
            obj.name: obj for name, obj in vars(module).items()
            if isinstance(obj, Command) and not name.startswith('_')
        }

        return cls(commands)

    @classmethod
    def get_module(cls, path):
        raise_does_not_exist = False

        if path.endswith('.py'):
            commands_module = abs_path(path)
            if not os.path.isfile(commands_module):
                raise_does_not_exist = True
                does_not_exist_message = 'Commands file does not exist: {commands_module}'
            else:
                spec = spec_from_file_location('commands', commands_module)
                module = module_from_spec(spec)
                spec.loader.exec_module(module)
        else:
            try:
                module = import_module(path)
            except ImportError:
                raise_does_not_exist = True
                does_not_exist_message = 'Commands module could not be imported: {commands_module}'

        if raise_does_not_exist:
            raise RunnerError(does_not_exist_message.format_map(locals()))

        return module

    def set_attrs(self, **attrs):
        """Set the given attributes on *all* commands in collection."""
        commands = tuple(self.values())
        for name, value in attrs.items():
            for command in commands:
                setattr(command, name, value)

    def set_default_args(self, default_args):
        """Set default args for commands in collection.

        Default args are used when the corresponding args aren't passed
        on the command line or in a direct call.

        """
        for name, args in default_args.items():
            command = self[name]
            command.default_args = default_args.get(command.name) or {}

    def __getitem__(self, name):
        commands = self.commands
        if name in commands:
            return commands[name]
        name = Command.normalize_name(name)
        return commands[name]

    def __setitem__(self, name, command):
        self.commands[name] = command

    def __delitem__(self, name):
        del self.commands[name]

    def __iter__(self):
        return iter(self.commands)

    def __len__(self):
        return len(self.commands)



import os
from unittest import TestCase

from ..collection import Collection
from ..command import command
from ..run import run
from ..runner import CommandRunner


CONFIG_FILE = os.path.join(os.path.dirname(__file__), 'commands.yaml')


@command
def test(a, b, c, d=None):
    return a, b, c, d


class TestConfig(TestCase):

    def setUp(self):
        self.collection = Collection({'test': test})

    def read_config_file(self, config_file=CONFIG_FILE):
        return run.read_config_file(config_file, self.collection)

    def interpolate(self, config):
        globals_, default_args, environ = run.interpolate(
            config.get('globals') or {},
            config.get('args') or {},
            config.get('environ') or {},
        )
        return {
            'globals': globals_,
            'args': default_args,
            'environ': environ,
        }

    def test_read_config(self):
        config = self.read_config_file()
        self.assertIn('globals', config)
        self.assertIn('args', config)
        self.assertIn('environ', config)
        self.assertIn('env', config['globals'])
        self.assertEqual('test', config['globals']['env'])

    def test_read_config_and_interpolate(self):
        config = self.read_config_file()
        config = self.interpolate(config)
        self.assertEqual({'env': 'test', 'a': 'b', 'b': 'b', 'd': 'd'}, config['globals'])
        self.assertEqual({'test': {'a': 'b', 'b': 'b', 'd': 'x'}}, config['args'])
        self.assertEqual({'XXX': 'b', 'XYZ': 'b'}, config['environ'])

    def test_read_config_then_call_command(self):
        config = self.read_config_file()
        config = self.interpolate(config)
        runner = CommandRunner(self.collection)
        self.collection.set_default_args(config['args'])

        # Uses default args
        result = runner.run(['test', 'c'])[0]
        self.assertEqual(('b', 'b', 'c', 'x'), result)

        # Uses some default args
        result = runner.run(['test', '--a', 'a', 'c'])[0]
        self.assertEqual(('a', 'b', 'c', 'x'), result)

        # Uses no default args
        result = runner.run(['test', '--a', 'x', '--b', 'y', 'c', '-d', 'z'])[0]
        self.assertEqual(('x', 'y', 'c', 'z'), result)



from .args import arg, bool_or, json_value
from .command import command, subcommand
from .util import abort, confirm, printer

__version__ = '1.0a54'

__all__ = [
    '__version__',
    'abort',
    'arg',
    'bool_or',
    'command',
    'confirm',
    'json_value',
    'printer',
    'subcommand',
]



import enum
import subprocess


class Color(enum.Enum):

    none = ''
    reset = '\033[0m'
    black = '\033[90m'
    red = '\033[91m'
    green = '\033[92m'
    yellow = '\033[93m'
    blue = '\033[94m'
    magenta = '\033[95m'
    cyan = '\033[96m'
    white = '\033[97m'

    def __str__(self):
        return self.value


class StreamOptions(enum.Enum):

    """Choices for stream handling."""

    capture = 'capture'
    hide = 'hide'
    none = 'none'

    @property
    def option(self):
        return {
            'capture': subprocess.PIPE,
            'hide': subprocess.DEVNULL,
            'none': None,
        }[self.value]



class RunCommandsError(Exception):

    pass


class RunAborted(RunCommandsError):

    def __init__(self, return_code=0, message='Aborted'):
        self.message = message
        self.return_code = return_code
        super().__init__(message, return_code)

    def __str__(self):
        return self.message


class RunnerError(RunCommandsError):

    pass


class CommandError(RunCommandsError):

    pass



import inspect
from collections import OrderedDict

from .decorators import cached_property
from .enums import Color, StreamOptions
from .misc import abort, flatten_args, format_if, isatty, load_object, merge_dicts
from .path import abs_path, asset_path, paths_to_str
from .printer import get_hr, printer
from .prompt import confirm, prompt
from .string import camel_to_underscore


__all__ = [
    'get_commands_in_namespace',
    'cached_property',
    'Color', 'StreamOptions',
    'abort', 'flatten_args', 'format_if', 'isatty', 'load_object', 'merge_dicts',
    'abs_path', 'asset_path', 'paths_to_str',
    'get_hr', 'printer',
    'confirm', 'prompt',
    'camel_to_underscore',
]


def get_commands_in_namespace(namespace=None, level=1):
    """Get commands in namespace.

    Args:
        namespace (dict|module): Typically a module. If not passed, the
            globals from the call site will be used.
        level (int): If not called from the global scope, set this
            appropriately to account for the call stack.

    Returns:
        OrderedDict: The commands found in the namespace, ordered by
            name.

    Can be used to create ``__all__`` lists::

        __all__ = list(get_commands_in_namespace())

    """
    from ..command import Command  # noqa: Avoid circular import
    commands = {}
    if namespace is None:
        frame = inspect.stack()[level][0]
        namespace = frame.f_globals
    elif inspect.ismodule(namespace):
        namespace = vars(namespace)
    for name in namespace:
        obj = namespace[name]
        if isinstance(obj, Command):
            commands[name] = obj
    return OrderedDict((name, commands[name]) for name in sorted(commands))



import argparse
import inspect
import signal
import sys
import time
from collections import OrderedDict
from inspect import Parameter
from typing import Mapping

from .args import POSITIONAL_PLACEHOLDER, Arg, ArgConfig, HelpArg
from .exc import CommandError, RunCommandsError
from .util import cached_property, camel_to_underscore, get_hr, printer


__all__ = ['command', 'subcommand', 'Command']


class Command:

    """Wraps a callable and provides a command line argument parser.

    Args:
        implementation (callable): A callable that implements the
            command's functionality. The command's console script will
            be generated by inspecting this callable.
        name (str): Name of command as it will be called from the
            command line. Defaults to ``implementation.__name__`` (with
            underscores replaced with dashes).
        description (str): Description of command shown in command
            help. Defaults to ``implementation.__doc__``.
        timed (bool): Whether the command should be timed. Will print an
            info message showing how long the command took to complete
            when ``True``. Defaults to ``False``.
        arg_config (dict): For commands defined as classes, this can be
            used to configure common base args instead of repeating the
            configuration for each subclass. Note that its keys should
            be actual parameter names and not normalized arg names.
        debug (bool): When this is set, additional debugging info will
            be shown.

    Other attributes:
        module: Module containing command
            For a class, this is just ``self.__class__.__module__``
            For a function, this is ``self.implementation.__module__``
        qualname: Qualified name of command within module
            For a class, this is just ``self.__class__.__qualname``
            For a function, this is ``self.implementation.__qualname__``

    This is typically used via the :func:`command` decorator::

        from runcommands import command

        @command
        def my_command():
            ...

    Decorating a function with :func:`command` will create an instance
    of :class:`Command` with the wrapped function as its implementation.

    Args can be passed to :func:`command`, which will be passed through
    to the :class:`Command` constructor::

        @command(name='better-name')
        def my_command():
            ...

    It's also possible to use a class directly as a command::

        @command
        class MyCommand(Command):

            def implementation(self):
                ...

    Using the :func:`command` decorator on a class will create an
    instance of the class in the namespace where the class is defined.

    Command Names:

    A command's name is derived from the normalized name of its
    implementation function by default::

        @command
        def some_command():
            ...

        # Command name: some-command

    A name can be set explicitly instead, in which case it *won't* be
    normalized::

        @command(name='do_stuff')
        def some_command():
            ...

        # Command name: do_stuff

    If the command is defined as a class, its name will be derived from
    its class name by default (split into words then normalized)::

        class SomeCommand(Command):
            ...

        # Command name: some-command

    The `command` decorator or a class-level attribute can be used to
    set the command's name explicitly::

        @command(name='do_stuff')
        class SomeCommand(Command):
            ...

        class SomeCommand(Command):
            name = 'do_stuff'

        # Command name in both cases: do_stuff

    """

    def __init__(self, implementation=None, name=None, description=None, base_command=None,
                 timed=False, arg_config=None, default_args=None, debug=False):
        if implementation is None:
            if not hasattr(self, 'implementation'):
                raise CommandError(
                    'Missing implementation; it must be passed in as a function or defined as a '
                    'method on the command class')
            self.module = self.__class__.__module__
            self.qualname = self.__class__.__qualname__
            default_name = self.normalize_class_name(self.__class__.__name__)
        else:
            self.implementation = implementation
            self.module = implementation.__module__
            self.qualname = implementation.__qualname__
            default_name = self.normalize_name(implementation.__name__)

        name = name or getattr(self.__class__, 'name', None) or default_name

        is_subcommand = base_command is not None

        if is_subcommand:
            name = ':'.join((base_command.name, name))

        description = description or self.get_description_from_docstring(self.implementation)
        short_description = description.splitlines()[0] if description else None

        self.name = name
        self.description = description
        self.short_description = short_description
        self.timed = timed
        self.arg_config = arg_config or {}
        self.debug = debug
        self.default_args = default_args or {}
        self.mutual_exclusion_groups = {}

        # Subcommand-related attributes
        first_arg = next(iter(self.args.values()), None)
        self.base_command = base_command
        self.is_subcommand = is_subcommand
        self.subcommands = []
        self.first_arg = first_arg
        self.first_arg_has_choices = False if first_arg is None else bool(first_arg.choices)

        if is_subcommand:
            base_command.add_subcommand(self)

    def subcommand(self, name=None, description=None, timed=False):
        """Create a subcommand of the specified base command."""
        base_command = self
        return command(name, description, base_command, timed, cls=self.__class__)

    @property
    def is_base_command(self):
        return bool(self.subcommands)

    @property
    def subcommand_depth(self):
        depth = 0
        base_command = self.base_command
        while base_command:
            depth += 1
            base_command = base_command.base_command
        return depth

    @cached_property
    def base_name(self):
        if self.is_subcommand:
            return self.name.split(':', self.subcommand_depth)[-1]
        return self.name

    @cached_property
    def prog_name(self):
        if self.is_subcommand:
            return ' '.join(self.name.split(':', self.subcommand_depth))
        return self.base_name

    def add_subcommand(self, subcommand):
        name = subcommand.base_name
        self.subcommands.append(subcommand)
        if not self.first_arg_has_choices:
            if self.first_arg.choices is None:
                self.first_arg.choices = []
            self.first_arg.choices.append(name)

    def get_description_from_docstring(self, implementation):
        description = implementation.__doc__
        if description is not None:
            description = description.strip() or None
        if description is not None:
            lines = description.splitlines()
            title = lines[0]
            if title.endswith('.'):
                title = title[:-1]
            lines = [title] + [line[4:] for line in lines[1:]]
            description = '\n'.join(lines)
        return description

    def run(self, argv=None, **overrides):
        if self.timed:
            start_time = time.monotonic()

        empty = Parameter.empty
        debug = self.debug
        positionals = self.positionals
        var_positional = self.var_positional
        default_args = self.default_args

        if argv is None:
            argv = sys.argv[1:]

        parsed_args = argv if isinstance(argv, dict) else self.parse_args(argv)

        args = []
        var_args = ()
        kwargs = parsed_args.copy()
        kwargs.update(overrides)

        if debug:
            # Names of all positional args.
            arg_names = []
            # Positional args passed via command line (name, value pairs).
            args_passed = []
            # Name of the var args arg.
            var_args_name = None
            # Positional args added from command's default args.
            from_default_args = {}
            # Positional args added from arg defaults.
            from_arg_defaults = {}

        # Map command line args to the command's parameters. Extract
        # positional args from the parsed args dict so they can be
        # passed positionally, using defaults for positionals that
        # weren't passed. Nothing special needs to be done for optional
        # args.
        for arg in positionals.values():
            name = arg.parameter.name
            value = kwargs.pop(name, POSITIONAL_PLACEHOLDER)
            if value is POSITIONAL_PLACEHOLDER:
                if name in default_args:
                    value = default_args[name]
                    if debug:
                        from_default_args[name] = value
                elif arg.default is not empty:
                    value = arg.default
                    if debug:
                        from_arg_defaults[name] = value
                else:
                    raise AssertionError('This should never happen')
            elif debug:
                args_passed.append((name, value))
            if debug:
                arg_names.append(name)
            args.append(value)

        if var_positional:
            var_args_name = var_positional.parameter.name
            if var_args_name in kwargs:
                var_args = kwargs.pop(var_args_name)
                if debug:
                    args_passed.append((var_args_name, var_args))

        if debug:
            printer.debug('Command called via command line:', self.name)
            printer.debug('    Parsed args:', parsed_args)
            printer.debug('    Overrides:', overrides)
            printer.debug('    Received positional args:', tuple(args_passed))
            printer.debug('    Received optional args:', kwargs)
            printer.debug('    Added positionals from default args:', ', '.join(from_default_args))
            printer.debug('    Added positionals from arg defaults:', ', '.join(from_arg_defaults))
            printer.debug('Running command via command line:', self.name)
            printer.debug('    Positional args:', tuple(zip(arg_names, args)))
            printer.debug('    Var args:', (var_args_name, var_args) if var_args else var_args)
            printer.debug('    Optional args:', kwargs)

        if var_args:
            args = tuple(args) + tuple(var_args)

        result = self(*args, **kwargs)

        if self.timed:
            self.print_elapsed_time(time.monotonic() - start_time)

        return result

    def console_script(self, argv=None, **overrides):
        debug = self.debug
        argv = sys.argv[1:] if argv is None else argv
        is_base_command = self.is_base_command

        if hasattr(self, 'sigint_handler'):
            signal.signal(signal.SIGINT, self.sigint_handler)

        if is_base_command:
            commands = self.partition_subcommands(argv)
        else:
            commands = [(self, argv)]

        for cmd, cmd_argv in commands:
            try:
                result = cmd.run(cmd_argv, **overrides)
            except RunCommandsError as result:
                if debug:
                    raise
                return_code = result.return_code if hasattr(result, 'return_code') else 1
                result_str = str(result)
                if result_str:
                    if return_code:
                        printer.error(result_str, file=sys.stderr)
                    else:
                        printer.print(result_str)
            else:
                return_code = result.return_code if hasattr(result, 'return_code') else 0

        return return_code

    def partition_subcommands(self, argv, base=True):
        debug = self.debug
        base_argv = []
        base_args = {}
        subcmd_args = {}
        commands = [(self, base_args)]
        subcommand_map = {sub.name: sub for sub in self.subcommands}

        if debug:
            printer.debug('Parsing command for subcommands:', self.name)
            printer.debug('    argv:', argv)

        for i, arg in enumerate(argv):
            if arg.startswith(':'):
                base_argv.append(arg[1:])
            else:
                base_argv.append(arg)
                qualified_name = '{self.name}:{arg}'.format_map(locals())

                if qualified_name in subcommand_map:
                    subcmd = subcommand_map[qualified_name]
                    remaining_argv = argv[i + 1:]

                    if debug:
                        printer.debug('Found subcommand:', subcmd.name)
                        printer.debug('    Base argv:', base_argv)
                        printer.debug('    Remaining argv:', remaining_argv)

                    base_args.update(self.parse_args(base_argv))

                    if subcmd.is_base_command:
                        commands.extend(subcmd.partition_subcommands(remaining_argv, False))
                    else:
                        subcmd_args = subcmd.parse_args(remaining_argv)
                        commands.append([subcmd, subcmd_args])

                    break
        else:
            # No subcommand found
            if debug:
                printer.debug('Found command with no subcommand:', self.name)
                printer.debug('    argv:', base_argv)
            base_args.update(self.parse_args(base_argv))

        if base:
            # Pass base args down to subcommands. Each subcommand will
            # receive args from *all* preceding base commands. The last
            # value for an arg will be used (i.e., if two base commands
            # have the same option and it was passed to both for some
            # reason).
            base_cmd, base_args = commands[0]
            base_args = base_args.copy()
            subcmd_arg_names = set(subcmd_args)
            for i, (subcmd, subcmd_args) in enumerate(commands[1:], 1):
                for name, value in base_args.items():
                    base_cmd_arg = base_cmd.find_arg(name)
                    subcmd_arg = subcmd.find_arg(name)
                    pass_down = (
                        # The subcommand has this option
                        subcmd_arg and
                        # And it wasn't passed directly to the subcommand
                        (name not in subcmd_arg_names) and
                        # And the base command has this arg too
                        base_cmd_arg and
                        # And it's optional...
                        base_cmd_arg.is_optional and
                        # ...but not an optional positional
                        (not base_cmd_arg.is_positional)
                    )
                    if pass_down:
                        subcmd_args[name] = value
                base_cmd = subcmd
                base_args.update(subcmd_args)
            if self.debug:
                printer.debug('Subcommands:')
                for cmd, cmd_argv in commands:
                    printer.debug('   ', cmd.name, cmd_argv)

        return commands

    def __call__(self, *passed_args, **passed_kwargs):
        empty = Parameter.empty
        debug = self.debug
        positionals = tuple(self.positionals.values())
        num_positionals = len(positionals)
        var_positional = self.var_positional
        default_args = self.default_args

        num_passed_args = len(passed_args)
        args = []
        var_args = ()
        kwargs = passed_kwargs.copy()

        if debug:
            # Positional args passed (name, value pairs).
            args_passed = []
            # Name of the var args arg.
            var_args_name = None
            # Args added from command's default args.
            from_default_args = {}
            # Args added from arg defaults.
            from_arg_defaults = {}

        # The N passed positional args are mapped to the first N
        # positional parameters.
        for arg, value in zip(positionals, passed_args):
            name = arg.parameter.name
            args.append((name, value))
            if debug:
                args_passed.append((name, value))

        # Use defaults for positionals that weren't passed. This is done
        # here instead of below with the optionals so they'll be passed
        # positionally.
        for arg in positionals[len(args):]:
            name = arg.parameter.name
            if name in default_args:
                value = default_args[name]
                args.append((name, value))
                if debug:
                    from_default_args[name] = value
            elif arg.default is not empty:
                value = arg.default
                args.append((name, value))
                if debug:
                    from_arg_defaults[name] = value

        # If the command has var args, it consumes any remaining passed
        # positional args.
        if var_positional:
            var_args_name = var_positional.parameter.name
            var_args = passed_args[len(args):]
            if var_args:
                if debug:
                    args_passed.append((var_args_name, var_args))
            elif var_args_name in default_args:
                var_args = default_args[var_args_name]
            elif var_positional.default is not empty:
                var_args = var_positional.default

        # Otherwise, the remaining N passed positionals args correspond
        # to the first N optionals.
        elif num_passed_args > num_positionals:
            for arg, value in zip(self.optionals.values(), passed_args[num_positionals:]):
                name = arg.name
                args.append((name, value))
                if debug:
                    args_passed.append((name, value))

        # Use defaults for any optionals that weren't passed that have a
        # default. Positionals that weren't passed have already had
        # their defaults set above.
        arg_names = tuple(item[0] for item in args)
        get_from_default_args = set(default_args).difference(arg_names, passed_kwargs)
        for name in get_from_default_args:
            value = default_args[name]
            kwargs[name] = value
            if debug:
                from_default_args[name] = value

        if debug:
            var_args_display = (var_args_name, tuple(var_args)) if var_args else ()
            printer.debug('Command called:', self.name)
            printer.debug('    Received positional args:', args_passed)
            printer.debug('    Received keyword args:', passed_kwargs)
            printer.debug('    Added from default args:', from_default_args)
            printer.debug('    Added from arg defaults:', from_arg_defaults)
            printer.debug('Running command:', self.name)
            printer.debug('    Positional args:', args)
            printer.debug('    Var args:', var_args_display)
            printer.debug('    Keyword args:', kwargs)

        args = tuple(item[1] for item in args)

        if var_args:
            args = args + tuple(var_args)

        return self.implementation(*args, **kwargs)

    def parse_args(self, argv, expand_short_options=True):
        if self.debug:
            printer.debug('Parsing args for command `{self.name}`: {argv}'.format_map(locals()))
        if expand_short_options:
            argv = self.expand_short_options(argv)
        parsed_args = self.arg_parser.parse_args(argv)
        parsed_args = vars(parsed_args)
        for k, v in parsed_args.items():
            if v == '':
                parsed_args[k] = None
        return parsed_args

    def parse_optional(self, string):
        """Parse string into name, option, and value (if possible).

        If the string is a known option name, the string, the
        corresponding option, and ``None`` will be returned.

        If the string has the form ``--option=<value>`` or
        ``-o=<value>``, it will be split on equals into an option name
        and value. If the option name is known, the option name, the
        corresponding option, and the value will be returned.

        In all other cases, ``None`` will be returned to indicate that
        the string doesn't correspond to a known option.

        """
        option_map = self.option_map
        if string in option_map:
            return string, option_map[string], None
        if '=' in string:
            name, value = string.split('=', 1)
            if name in option_map:
                return name, option_map[name], value
        return None

    def expand_short_options(self, argv):
        """Convert grouped short options like `-abc` to `-a, -b, -c`.

        This is necessary because we set ``allow_abbrev=False`` on the
        ``ArgumentParser`` in :attr:`self.arg_parser`. The argparse docs
        say ``allow_abbrev`` applies only to long options, but it also
        affects whether short options grouped behind a single dash will
        be parsed into multiple short options.

        See :meth:`parse_multi_short_option` for details on how multi
        short options are parsed.

        Returns:
            list: Original argv if no multi short options found
            list: Expanded argv if multi short options found

        """
        if self.debug:
            printer.debug('Expanding short options for `{self.name}`: {argv}'.format_map(locals()))
        has_multi_short_options = False
        parse_multi_short_option = self.parse_multi_short_option
        new_argv = []
        for i, arg in enumerate(argv):
            if arg == '--':
                new_argv.extend(argv[i:])
                break
            short_options, value = parse_multi_short_option(arg)
            if short_options:
                new_argv.extend(short_options)
                if value is not None:
                    new_argv.append(value)
                has_multi_short_options = True
            else:
                new_argv.append(arg)
        if self.debug and not has_multi_short_options:
            printer.debug('No multi short options found')
        return new_argv if has_multi_short_options else argv

    def parse_multi_short_option(self, arg):
        """Parse args like '-xyz' into ['-x', '-y', '-z'].

        Examples::

            'abc' -> None, None                   # not an option
            '--option' -> None, None              # long option
            '-a' -> None, None                    # short option but not multi
            '-xyz' -> ['-x', '-y', '-z'], None    # multi short option
            '-xyz12' -> ['-x', '-y', '-z'], '12'  # multi short option w/ value

        Note that parsing stops when a short option that takes a value
        is encountered--the rest of the arg string is considered the
        value for that option.

        Returns:
            (None, None): The arg is not a multi short option
            (list, str|None): The arg is a multi short option (perhaps
                including a value for the last option)

        """
        if len(arg) < 3 or arg[0] != '-' or arg[1] == '-' or arg[2] == '=':
            # Not a multi short option like '-abc'.
            return None, None
        # Appears to be a multi short option.
        option_map = self.option_map
        short_options = []
        value = None
        for i, char in enumerate(arg[1:], 1):
            name = '-{char}'.format(char=char)
            short_options.append(name)
            option = option_map.get(name)
            if option is not None and option.takes_value:
                j = i + 1
                if j < len(arg):
                    value = arg[j:]
                break
        if self.debug and short_options:
            printer.debug('Parsed multi short option:', arg, '=>', short_options)
        return short_options, value

    @staticmethod
    def normalize_name(name):
        # Chomp a single trailing underscore *if* the name ends with
        # just one trailing underscore. This accommodates the convention
        # of adding a trailing underscore to reserved/built-in names.
        if name.endswith('_'):
            if name[-2] != '_':
                name = name[:-1]
        name = name.replace('_', '-')
        return name

    @staticmethod
    def normalize_class_name(name):
        name = camel_to_underscore(name)
        name = name.replace('_', '-')
        name = name.lower()
        return name

    def find_arg(self, name):
        """Find arg by normalized arg name or parameter name."""
        name = self.normalize_name(name)
        return self.args.get(name)

    def find_parameter(self, name):
        """Find parameter by name or normalized arg name."""
        name = self.normalize_name(name)
        arg = self.args.get(name)
        return None if arg is None else arg.parameter

    def get_arg_config(self, param):
        annotation = param.annotation
        if annotation is param.empty:
            annotation = self.arg_config.get(param.name) or ArgConfig()
        elif isinstance(annotation, type):
            annotation = ArgConfig(type=annotation)
        elif isinstance(annotation, str):
            annotation = ArgConfig(help=annotation)
        elif isinstance(annotation, Mapping):
            annotation = ArgConfig(**annotation)
        return annotation

    def get_short_option_for_arg(self, name, used):
        first_char = name[0]
        first_char_upper = first_char.upper()

        if name == 'help':
            candidates = (first_char,)
        elif name.startswith('h'):
            candidates = (first_char_upper,)
        else:
            candidates = (first_char, first_char_upper)

        for char in candidates:
            short_option = '-{char}'.format_map(locals())
            if short_option not in used:
                return short_option

    def get_long_option_for_arg(self, name):
        return '--{name}'.format_map(locals())

    def get_inverse_short_option_for_arg(self, short_option, used):
        inverse_short_option = short_option.upper()
        if inverse_short_option not in used:
            return inverse_short_option

    def get_inverse_long_option_for_arg(self, long_option):
        if long_option == '--yes':
            return '--no'
        if long_option == '--no':
            return '--yes'
        if long_option.startswith('--no-'):
            return long_option.replace('--no-', '--', 1)
        if long_option.startswith('--is-'):
            return long_option.replace('--is-', '--not-', 1)
        if long_option.startswith('--with-'):
            return long_option.replace('--with-', '--without-', 1)
        return long_option.replace('--', '--no-', 1)

    def print_elapsed_time(self, elapsed_time):
        m, s = divmod(elapsed_time, 60)
        m = int(m)
        hr = get_hr()
        msg = '{hr}\nElapsed time for {self.name} command: {m:d}m {s:.3f}s\n{hr}'
        msg = msg.format_map(locals())
        printer.info(msg)

    @cached_property
    def parameters(self):
        implementation = self.implementation
        signature = inspect.signature(implementation)
        return signature.parameters

    @cached_property
    def has_kwargs(self):
        return any(p.kind is p.VAR_KEYWORD for p in self.parameters.values())

    @cached_property
    def args(self):
        """Create args from function parameters."""
        params = self.parameters
        args = OrderedDict()

        empty = Parameter.empty
        keyword_only = Parameter.KEYWORD_ONLY
        var_keyword = Parameter.VAR_KEYWORD
        var_positional = Parameter.VAR_POSITIONAL

        normalize_name = self.normalize_name
        get_arg_config = self.get_arg_config
        get_short_option = self.get_short_option_for_arg
        get_long_option = self.get_long_option_for_arg
        get_inverse_short_option = self.get_inverse_short_option_for_arg
        get_inverse_long_option = self.get_inverse_long_option_for_arg

        params = OrderedDict((
            (normalize_name(n), p)
            for n, p in params.items()
            if not (
                (n.startswith('_')) or
                (p.kind is keyword_only and p.default is empty) or
                (p.kind is var_keyword)
            )
        ))

        used_short_options = set()
        for param in params.values():
            annotation = get_arg_config(param)
            short_option = annotation.short_option
            if short_option:
                used_short_options.add(short_option)

        for name, param in params.items():
            annotation = get_arg_config(param)
            container = annotation.container
            type = annotation.type
            choices = annotation.choices
            help = annotation.help
            inverse_help = annotation.inverse_help
            short_option = annotation.short_option
            long_option = annotation.long_option
            no_inverse = annotation.no_inverse
            inverse_short_option = annotation.inverse_short_option
            inverse_long_option = annotation.inverse_long_option
            action = annotation.action
            nargs = annotation.nargs
            mutual_exclusion_group = annotation.mutual_exclusion_group

            default = param.default
            is_var_positional = param.kind is var_positional
            is_positional = default is empty and not is_var_positional

            if annotation.default is not empty:
                if is_positional or is_var_positional:
                    default = annotation.default
                else:
                    message = (
                        'Got default for `{self.name}` command\'s optional arg `{name}` via '
                        'arg annotation. Optional args must specify their defaults via keyword '
                        'arg values.'
                    ).format_map(locals())
                    raise CommandError(message)

            if not (is_positional or is_var_positional):
                if not short_option:
                    short_option = get_short_option(name, used_short_options)
                    used_short_options.add(short_option)
                if not long_option:
                    long_option = get_long_option(name)
                if not no_inverse:
                    if short_option and not inverse_short_option:
                        inverse_short_option = get_inverse_short_option(
                            short_option,
                            used_short_options,
                        )
                        used_short_options.add(inverse_short_option)
                    if not inverse_long_option:
                        inverse_long_option = get_inverse_long_option(long_option)

            args[name] = Arg(
                command=self,
                parameter=param,
                name=name,
                container=container,
                type=type,
                positional=is_positional,
                default=default,
                choices=choices,
                help=help,
                inverse_help=inverse_help,
                short_option=short_option,
                long_option=long_option,
                no_inverse=no_inverse,
                inverse_short_option=inverse_short_option,
                inverse_long_option=inverse_long_option,
                action=action,
                nargs=nargs,
                mutual_exclusion_group=mutual_exclusion_group,
            )

        if 'help' not in args:
            args['help'] = HelpArg(command=self)

        option_map = OrderedDict()
        for arg in args.values():
            for option in arg.options:
                option_map.setdefault(option, [])
                option_map[option].append(arg)

        for option, option_args in option_map.items():
            if len(option_args) > 1:
                names = ', '.join(a.parameter.name for a in option_args)
                message = (
                    'Option {option} of command {self.name} maps to multiple parameters: {names}')
                message = message.format_map(locals())
                raise CommandError(message)

        return args

    @cached_property
    def arg_parser(self):
        use_default_help = isinstance(self.args['help'], HelpArg)

        parser = argparse.ArgumentParser(
            prog=self.prog_name,
            description=self.description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            argument_default=argparse.SUPPRESS,
            add_help=use_default_help,
            allow_abbrev=False,  # See note in `self.parse_args()`
        )

        default_args = self.default_args

        for name, arg in self.args.items():
            if name == 'help' and use_default_help:
                continue

            param = arg.parameter
            options, kwargs = arg.add_argument_args

            if arg.is_positional and param.name in default_args:
                # Positionals are made optional if a default value is
                # specified via config.
                options = (self.get_long_option_for_arg(arg.name),)

            mutual_exclusion_group_name = arg.mutual_exclusion_group
            if mutual_exclusion_group_name:
                if mutual_exclusion_group_name not in self.mutual_exclusion_groups:
                    self.mutual_exclusion_groups[mutual_exclusion_group_name] = \
                        parser.add_mutually_exclusive_group()
                mutual_exclusion_group = self.mutual_exclusion_groups[mutual_exclusion_group_name]
                mutual_exclusion_group.add_argument(*options, **kwargs)
            else:
                parser.add_argument(*options, **kwargs)

            inverse_args = arg.add_argument_inverse_args
            if inverse_args is not None:
                options, kwargs = inverse_args
                parser.add_argument(*options, **kwargs)

        return parser

    @cached_property
    def positionals(self):
        args = self.args.items()
        return OrderedDict((name, arg) for (name, arg) in args if arg.is_positional)

    @cached_property
    def var_positional(self):
        args = self.args.items()
        for name, arg in args:
            if arg.is_var_positional:
                return arg
        return None

    @cached_property
    def optionals(self):
        args = self.args.items()
        return OrderedDict((name, arg) for (name, arg) in args if arg.is_optional)

    @cached_property
    def option_map(self):
        """Map command-line options to args."""
        option_map = OrderedDict()
        for arg in self.args.values():
            for option in arg.options:
                option_map[option] = arg
        return option_map

    @property
    def help(self):
        help_ = self.arg_parser.format_help()
        help_ = help_.split(': ', 1)[1]
        help_ = help_.strip()
        return help_

    @property
    def usage(self):
        usage = self.arg_parser.format_usage()
        usage = usage.split(': ', 1)[1]
        usage = usage.strip()
        return usage

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.usage

    def __repr__(self):
        return 'Command(name={self.name})'.format(self=self)


def command(name=None, description=None, base_command=None, timed=False, cls=Command):
    args = dict(description=description, base_command=base_command, timed=timed)

    if isinstance(name, type):
        # Bare class decorator
        name.implementation.__name__ = camel_to_underscore(name.__name__)
        return name(**args)

    if callable(name):
        # Bare function decorator
        return cls(implementation=name, **args)

    def wrapper(wrapped):
        if isinstance(wrapped, type):
            wrapped.implementation.__name__ = camel_to_underscore(wrapped.__name__)
            return wrapped(name=name, **args)
        return cls(implementation=wrapped, name=name, **args)

    return wrapper


def subcommand(base_command, name=None, description=None, timed=False, cls=Command):
    return command(name, description, base_command, timed, cls)



from .. import arg, bool_or
from ..command import command
from .local import local


@command
def git_version(
    short: arg(
        type=bool_or(int),
        help='Get short hash; optionally specify minimum length of hash',
    ) = True,
    show: 'Print version to stdout' = False
):
    """Get tag associated with HEAD; fall back to SHA1.

    If HEAD is tagged, return the tag name; otherwise fall back to
    HEAD's short SHA1 hash.

    .. note:: Only annotated tags are considered.

    .. note:: When no minimum hash length is specified, the minimum
        length is determined by ``git rev-parse`` based on git's
        ``core.abbrev`` config variable.

    .. note:: The output isn't shown by default. To show it, pass the
        ``--show`` flag.

    """
    result = local(
        ['git', 'rev-parse', '--is-inside-work-tree'],
        stdout='hide', stderr='hide', echo=False, raise_on_error=False)

    if not result:
        # Not a git directory
        return None

    # Return a tag if possible
    result = local(
        ['git', 'describe', '--exact-match'],
        stdout='capture', stderr='hide', echo=False, raise_on_error=False)

    if not result:
        # Fall back to hash
        args = ['git', 'rev-parse']
        if short:
            arg = '--short' if isinstance(short, bool) else '--short={short}'.format_map(locals())
            args.append(arg)
        args.append('HEAD')
        result = local(args, stdout='capture', stderr='hide', echo=False, raise_on_error=False)

    if result:
        version = result.stdout.strip()
        if show:
            print(version)
        return version

    return None



import ast
import os
import sys
import yaml

from jinja2 import Environment as TemplateEnvironment, TemplateRuntimeError

from . import __version__
from .args import arg, json_value
from .command import Command
from .collection import Collection
from .const import DEFAULT_COMMANDS_MODULE
from .exc import RunnerError
from .runner import CommandRunner
from .util import abs_path, merge_dicts, printer


class Run(Command):

    name = 'runcommands'

    allowed_config_file_args = (
        'globals',
        'envs',
        'args',
        'environ',
    )

    def implementation(self,
                       commands_module: arg(short_option='-m') = DEFAULT_COMMANDS_MODULE,
                       config_file: arg(short_option='-f') = None,
                       # Globals
                       globals_: arg(
                           container=dict,
                           type=json_value,
                           help='Global variables & default args for *all* commands; will be '
                                'injected into itself, default args, and environment variables '
                                '(higher precedence than keyword args)'
                       ) = None,
                       # Special globals (for command line convenience)
                       env: arg(help='env will be added to globals if specified') = None,
                       version: arg(help='version will be added to globals if specified') = None,
                       echo: arg(
                           type=bool,
                           help='echo=True will be added to globals',
                           inverse_help='echo=False will be added to globals'
                       ) = None,
                       # Environment variables
                       environ: arg(
                           container=dict,
                           help='Additional environment variables; '
                                'added just before commands are run'
                       ) = None,
                       # Meta
                       info: arg(help='Show info and exit') = False,
                       list_commands: arg(help='Show info & commands and exit') = False,
                       debug: arg(
                           type=bool,
                           help='Print debugging info & re-raise exceptions; also added to globals'
                       ) = None,
                       *,
                       all_argv=(),
                       run_argv=(),
                       command_argv=(),
                       cli_args=()):
        """Run one or more commands in succession.

        For example, assume the commands ``local`` and ``remote`` have been
        defined; the following will run ``ls`` first on the local host and
        then on the remote host::

            runcommands local ls remote <host> ls

        When a command name is encountered in ``argv``, it will be considered
        the starting point of the next command *unless* the previous item in
        ``argv`` was an option like ``--xyz`` that expects a value (i.e.,
        it's not a flag).

        To avoid ambiguity when an option value matches a command name, the
        value can be prepended with a colon to force it to be considered
        a value and not a command name.

        """
        collection = Collection.load_from_module(commands_module)
        config_file = self.find_config_file(config_file)
        cli_globals = globals_ or {}

        if env:
            cli_globals['env'] = env
        if version:
            cli_globals['version'] = version
        if echo is not None:
            cli_globals['echo'] = echo
        if debug is not None:
            cli_globals['debug'] = debug

        if config_file:
            args_from_file = self.read_config_file(config_file, collection)
            args = merge_dicts(args_from_file, {'environ': environ or {}})
            config_file_globals = args['globals']
            env = env or config_file_globals.get('env')

            if env:
                envs = args['envs']

                try:
                    env_globals = envs[env]
                except KeyError:
                    raise RunnerError('Unknown env: {env}'.format_map(locals()))

                # Don't add the selected env's default args dict to
                # globals, but save it so it can be added back to the
                # the global envs dict (for inspection purposes).
                env_default_args = env_globals.pop('args')

                globals_ = merge_dicts(config_file_globals, env_globals, cli_globals)
                globals_['envs'] = envs

                env_globals['args'] = env_default_args
            else:
                env_default_args = {}
                globals_ = merge_dicts(config_file_globals, cli_globals)

            base_default_args = merge_dicts(args['args'], env_default_args)

            default_args = {name: {} for name in collection}
            default_args = merge_dicts(default_args, base_default_args)

            for command_name, command_default_args in default_args.items():
                command = collection[command_name]

                # Add globals that correspond to this command (that
                # aren't present in default args section).
                for name, value in globals_.items():
                    param = command.find_parameter(name)
                    if param is not None:
                        if param.name not in command_default_args:
                            command_default_args[param.name] = value
                    elif command.has_kwargs:
                        name = Command.normalize_name(name)
                        command_default_args[name] = value

                if 'default_args' not in default_args:
                    # This gives commands access to both their own and
                    # other commands' default args.
                    if command.find_parameter('default_args'):
                        command_default_args['default_args'] = base_default_args

                # Convert lists to tuples for the command's args that are
                # specified as being tuples.
                for name, value in command_default_args.items():
                    command_arg = command.find_arg(name)
                    if command_arg and command_arg.container and isinstance(value, list):
                        command_default_args[name] = command_arg.container(value)

            default_args = {name: args for name, args in default_args.items() if args}

            environ = args['environ']
        else:
            globals_ = cli_globals
            default_args = {}
            environ = environ or {}

        debug = globals_.get('debug', False)
        show_info = info or list_commands or not command_argv or debug
        print_and_exit = info or list_commands

        globals_, default_args, environ = self.interpolate(globals_, default_args, environ)

        if show_info:
            print('RunCommands', __version__)

        if debug:
            printer.debug('Commands module:', commands_module)
            printer.debug('Config file:', config_file)
            printer.debug('All args:', all_argv)
            printer.debug('Run args:', run_argv)
            printer.debug('Command args:', command_argv)
            items = (
                ('Globals:', globals_),
                ('Default args:', default_args),
                ('Env default args:', env_default_args),
                ('Environment variables:', environ),
            )
            for label, data in items:
                if data:
                    printer.debug(label)
                    for k in sorted(data):
                        v = data[k]
                        printer.debug('  - {k} = {v!r}'.format_map(locals()))
                else:
                    printer.debug(label, data)

        if environ:
            os.environ.update(environ)

        collection.set_attrs(debug=debug)
        collection.set_default_args(default_args)
        runner = CommandRunner(collection, debug)

        if print_and_exit:
            if list_commands:
                runner.print_usage()
        elif not command_argv:
            printer.warning('\nNo command(s) specified')
            runner.print_usage()
        else:
            runner.run(command_argv)

    def run(self, argv, **kwargs):
        all_argv, run_argv, command_argv = self.partition_argv(argv)
        if '-d' in run_argv or '--debug' in run_argv:
            self.debug = True
        cli_argv = self.parse_args(run_argv, False)
        kwargs.update({
            'all_argv': all_argv,
            'run_argv': run_argv,
            'command_argv': command_argv,
            'cli_args': tuple(cli_argv),
        })
        return super().run(cli_argv, **kwargs)

    def partition_argv(self, argv=None):
        if argv is None:
            argv = sys.argv[1:]

        if not argv:
            return argv, [], []

        # Consume all args that appear to be options (and their values,
        # if applicable), even those that aren't know run options, up
        # until the first non-option word is reached. That word is
        # assumed to be the start of commands.

        def looks_like_option(s):
            return bool(
                (s.startswith('-') or s.startswith('--')) and
                not s.startswith('---') and
                s.strip('-')
            )

        i = 0
        argc = len(argv)
        run_argv = []
        parse_optional = self.parse_optional
        parse_multi_short_option = self.parse_multi_short_option

        while i < argc:
            a = argv[i]

            if a == '--':
                # Explicit end of run args.
                i += 1
                break

            option_data = parse_optional(a)

            if option_data is not None:
                # Arg is a known run option.
                name, option, value = option_data
                run_argv.append(a)

                if a in ('-d', '--debug'):
                    self.debug = True
                elif value is None and option.takes_value:
                    # Collect the option's value if it takes one and one
                    # wasn't provided via --opt=<value>.
                    j = i + 1
                    if j < argc:
                        run_argv.append(argv[j])
                        i = j
            else:
                if not looks_like_option(a):
                    # Non-option word; assumed to be start of commands.
                    break

                short_options, value = parse_multi_short_option(a)

                if short_options is None:
                    run_argv.append(a)
                else:
                    run_argv.extend(short_options)

                    if '-d' in short_options:
                        self.debug = True

                    if value is not None:
                        run_argv.append(value)
                    else:
                        # Collect the last short option's value if it
                        # takes one and one wasn't provided via
                        # -abc<value>.
                        option_data = parse_optional(short_options[-1])
                        if option_data is not None:
                            name, option, value = option_data
                            if value is None and option.takes_value:
                                j = i + 1
                                if j < argc:
                                    run_argv.append(argv[j])
                                    i = j

            i += 1

        command_argv = argv[i:]
        return argv, run_argv, command_argv

    def find_config_file(self, config_file):
        if config_file:
            config_file = abs_path(config_file)
        elif os.path.exists('commands.yaml'):
            config_file = abs_path('commands.yaml')
        return config_file

    def read_config_file(self, config_file, collection):
        return self._read_config_file(config_file, collection)

    def _read_config_file(self, config_file, collection):
        with open(config_file) as fp:
            args = yaml.load(fp, Loader=yaml.FullLoader) or {}

        for name in self.allowed_config_file_args:
            # Not present or present but not set
            if args.get(name) is None:
                args[name] = {}

        extends = args.pop('extends', None)

        for name in tuple(args):
            if name not in self.allowed_config_file_args:
                raise RunnerError(
                    'Arg cannot be specified in config file: {name}'
                    .format_map(locals()))

        self.normalize_command_and_arg_names(args['args'], collection)

        envs = args['envs']
        for env, data in envs.items():
            if data is None:
                data = {}
                envs[env] = data
            data.setdefault('args', {})
            self.normalize_command_and_arg_names(data['args'], collection)

        if extends:
            extends = abs_path(extends, relative_to=os.path.dirname(config_file))
            extended_args = self._read_config_file(extends, collection)
            args = merge_dicts(extended_args, args)

        return args

    def normalize_command_and_arg_names(self, config, collection):
        for command_name in tuple(config):
            try:
                command = collection[command_name]
            except KeyError:
                raise RunnerError(
                    'Unknown command in default args section of {config_file}: {command_name}'
                    .format_map(locals()))

            if command.name != command_name:
                config[command.name] = config.pop(command_name)

            command_default_args = config[command.name]
            for name in tuple(command_default_args):
                param = command.find_parameter(name)
                if param is None:
                    raise RunnerError(
                        'Unknown arg for command {command_name} in default args section of '
                        '{config_file}: {name}'.format_map(locals()))
                if param is not None and name != param.name:
                    command_default_args[param.name] = command_default_args.pop(name)

    def interpolate(self, globals_, default_args, environ):
        environment = TemplateEnvironment()

        if globals_:
            globals_ = self._interpolate(environment, globals_, globals_)

        if default_args:
            context = merge_dicts(globals_, default_args)
            default_args = self._interpolate(environment, default_args, context)

        if environ:
            environ = self._interpolate(environment, environ, globals_)

        return globals_, default_args, environ

    def _interpolate(self, environment, obj, context):
        if isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = self._interpolate(environment, v, context)
        elif isinstance(obj, list):
            for i, v in enumerate(obj):
                obj[i] = self._interpolate(environment, v, context)
        elif isinstance(obj, tuple):
            obj = tuple(self._interpolate(environment, v, context) for v in obj)
        elif isinstance(obj, str):
            while True:
                if '{{' not in obj:
                    break
                original_obj = obj
                template = environment.from_string(obj)
                try:
                    obj = template.render(context)
                except TemplateRuntimeError as exc:
                    raise RunnerError(
                        'Could not render template {obj!r} with context {context!r}: {exc}'
                        .format(exc=exc))
                if obj == original_obj:
                    break
            try:
                obj = ast.literal_eval(obj)
            except (SyntaxError, ValueError):
                pass
        return obj

    def sigint_handler(self, _sig_num, _frame):
        printer.warning('Aborted')
        sys.exit(0)


run = Run()



import argparse
import builtins
import json
import re
from enum import Enum
from inspect import Parameter
from typing import Mapping, Sequence

from .exc import CommandError
from .util import cached_property


EMPTY = Parameter.empty
KEYWORD_ONLY = Parameter.KEYWORD_ONLY
POSITIONAL_OR_KEYWORD = Parameter.POSITIONAL_OR_KEYWORD
VAR_POSITIONAL = Parameter.VAR_POSITIONAL


class POSITIONAL_PLACEHOLDER:

    """Used as a placeholder for positionals."""


class ArgConfig:

    """Configuration for an arg.

    This can be used as a function parameter annotation to explicitly
    configure an arg, overriding default behavior.

    Args:

        container (type): Container type to collect args into. If this
            is passed, or if the ``default`` value is a container type,
            values for this arg will be collected into a container of
            the appropriate type.
        type (type): The arg's type. By default, a positional arg will
            be parsed as ``str`` and an optional/keyword arg will be
            parsed as the type of its default value (or as ``str`` if
            the default value is ``None``). If a ``container`` is
            specified, or if the ``default`` value for the arg is a
            container, the ``type`` will be applied to the container's
            values.
        choices (sequence): A sequence of allowed choices for the arg.
        help (str): Help string for the arg.
        inverse_help (str): Inverse help string for the arg (for the
            ``--no-xyz`` variation of boolean args).
        short_option (str): Short command line option.
        long_option (str): Long command line option.
        inverse_short_option (str): Inverse short option for boolean
            args.
        inverse_long_option (str): Inverse long option for boolean args.
        action (Action): ``argparse`` Action.
        nargs (int|str): Number of command line args to consume.
        mutual_exclusion_group (str): Name of mutual exclusion group to
            add this arg to.
        default: Default value for positional args.

    .. note:: For convenience, regular dicts can be used to annotate
        args instead; they will be converted to instances of this class
        automatically.

    """

    short_option_regex = re.compile(r'-\w')
    long_option_regex = re.compile(r'--\w+(-\w+)*')

    def __init__(self, *,
                 container=None,
                 type=None,
                 choices=None,
                 help=None,
                 inverse_help=None,
                 short_option=None,
                 long_option=None,
                 no_inverse=False,
                 inverse_short_option=None,
                 inverse_long_option=None,
                 inverse_option=None,  # XXX: Temporary alias for inverse_long_option
                 action=None,
                 nargs=None,
                 mutual_exclusion_group=None,
                 default=EMPTY):
        if short_option is not None:
            if not self.short_option_regex.fullmatch(short_option):
                message = 'Expected short option with form -x, not "{short_option}"'
                message = message.format_map(locals())
                raise CommandError(message)

        if long_option is not None:
            if not self.long_option_regex.fullmatch(long_option):
                message = 'Expected long option with form --option, not "{long_option}"'
                message = message.format_map(locals())
                raise CommandError(message)

        self.container = container
        self.type = type
        self.choices = choices
        self.help = help
        self.inverse_help = inverse_help
        self.short_option = short_option
        self.long_option = long_option
        self.no_inverse = no_inverse
        self.inverse_short_option = inverse_short_option
        self.inverse_long_option = inverse_long_option or inverse_option
        self.action = action
        self.nargs = nargs
        self.mutual_exclusion_group = mutual_exclusion_group
        self.default = default

    def __repr__(self):
        options = (
            self.short_option,
            self.long_option,
            self.inverse_short_option,
            self.inverse_long_option,
        )
        options = (option for option in options if option)
        options = ', '.join(options)
        return 'arg<{self.type.__name__}>({options})'.format_map(locals())


arg = ArgConfig


class Arg:

    """Encapsulates an arg belonging to a command.

    Attributes:

        command (Command): Command this arg belongs to.
        parameter (Parameter): Function parameter this arg is derived
            from.
        name (str): Normalized arg name.
        container (type): Container type to collect args into. If this
            is passed, or if the ``default`` value is a container type,
            values for this arg will be collected into a container of
            the appropriate type.
        type (type): The arg's type. By default, a positional arg will
            be parsed as ``str`` and an optional/keyword arg will be
            parsed as the type of its default value (or as ``str`` if
            the default value is ``None``). If a ``container`` is
            specified, or if the ``default`` value for the arg is a
            container, the ``type`` will be applied to the container's
            values.
        positional (bool): An arg will automatically be considered
            positional if it doesn't have a default value, so this
            doesn't usually need to be passed explicitly. It can be
            used to force an arg that would normally be optional to
            be positional.
        default (object): Default value for the arg.
        choices (sequence): A sequence of allowed choices for the arg.
        help (str): Help string for the arg.
        inverse_help (str): Inverse help string for the arg (for the
            ``--no-xyz`` variation of boolean args).
        short_option (str): Short command line option.
        long_option (str): Long command line option.
        inverse_short_option (str): Inverse short option for boolean
            args.
        inverse_long_option (str): Inverse long option for boolean args.
        action (Action): ``argparse`` Action.
        nargs (int|str): Number of command line args to consume.
        mutual_exclusion_group (str): Name of mutual exclusion group to
            add this arg to.

    """

    def __init__(self, *,
                 command,
                 parameter,
                 name,
                 container,
                 type,
                 positional,
                 default,
                 choices,
                 help,
                 inverse_help,
                 short_option,
                 long_option,
                 no_inverse,
                 inverse_short_option,
                 inverse_long_option,
                 action,
                 nargs,
                 mutual_exclusion_group):

        command = command

        is_keyword_only = parameter.kind is KEYWORD_ONLY and default is EMPTY
        is_var_positional = parameter.kind is VAR_POSITIONAL
        if default is EMPTY:
            is_positional = not (is_var_positional or is_keyword_only)
            is_optional = False
        else:
            is_positional = False
            is_optional = True

        if positional is not None:
            is_positional = positional

        metavar = name.upper().replace('-', '_')
        if container and len(name) > 1 and name.endswith('s'):
            metavar = metavar[:-1]

        if container is None:
            is_mapping = isinstance(default, Mapping)
            is_sequence = isinstance(default, Sequence) and not isinstance(default, str)
            if is_mapping or is_sequence:
                container = default.__class__
            elif is_var_positional:
                container = tuple

        if type is None:
            if isinstance(choices, builtins.type) and issubclass(choices, Enum):
                type = choices
            elif container is None:
                if default not in (None, EMPTY):
                    type = default.__class__
                else:
                    type = str
            else:
                type = str

        if isinstance(type, builtins.type):
            is_bool = issubclass(type, bool)
            is_bool_or = issubclass(type, bool_or)
            is_enum_bool_or = is_bool_or and issubclass(type, Enum)
            is_enum = issubclass(type, Enum)
        else:
            is_bool = False
            is_bool_or = False
            is_enum_bool_or = False
            is_enum = False

        if is_bool:
            type = None
            metavar = None
        elif is_bool_or:
            type = type.type

        if not choices:
            if is_enum:
                choices = type
            elif is_enum_bool_or:
                choices = type.type

        if is_positional or is_var_positional:
            options = (short_option, long_option, inverse_long_option)
            options = tuple(option for option in options if option is not None)
            if options:
                raise CommandError(
                    'Positional args cannot be specified with options: {options}'
                    .format(options=', '.join(options)))

        if action is None:
            if container:
                action = ContainerAction.from_container(container)
            elif is_bool:
                action = 'store_true'
            elif is_bool_or:
                action = BoolOrAction

        if nargs is None:
            if is_positional:
                if container:
                    nargs = '+'
                elif is_optional:
                    nargs = '?'
            elif is_var_positional:
                nargs = '*'
            elif is_bool_or:
                nargs = '?'
            elif is_optional:
                if container:
                    nargs = '*'

        options = tuple(opt for opt in (short_option, long_option) if opt is not None)
        all_options = options

        if no_inverse:
            inverse_options = ()
        else:
            inverse_options = tuple(
                opt for opt in (inverse_short_option, inverse_long_option) if opt is not None
            )
            all_options += inverse_options

        if is_var_positional and default is EMPTY:
            default = ()

        self.command = command
        self.parameter = parameter
        self.is_positional = is_positional
        self.is_var_positional = is_var_positional
        self.is_optional = is_optional
        self.takes_value = is_positional or (is_optional and not is_bool)
        self.dest = parameter.name
        self.name = name
        self.metavar = metavar
        self.container = container
        self.type = type
        self.is_bool = is_bool
        self.is_bool_or = is_bool_or
        self.default = default
        self.choices = choices
        self.help = help
        self.inverse_help = inverse_help
        self.short_option = short_option
        self.long_option = long_option
        self.options = options
        self.no_inverse = no_inverse
        self.inverse_short_option = inverse_short_option
        self.inverse_long_option = inverse_long_option
        self.inverse_options = inverse_options
        self.all_options = all_options
        self.action = action
        self.nargs = nargs
        self.mutual_exclusion_group = mutual_exclusion_group

    @cached_property
    def add_argument_args(self):
        args = self.options
        kwargs = {
            'action': self.action,
            'choices': self.choices,
            'dest': self.dest,
            'help': self.help,
            'metavar': self.metavar,
            'nargs': self.nargs,
            'type': self.type,
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if self.is_positional and self.is_optional:
            kwargs['default'] = POSITIONAL_PLACEHOLDER
        return args, kwargs

    @cached_property
    def add_argument_inverse_args(self):
        if not (self.is_bool or self.is_bool_or) or self.no_inverse:
            return None

        args = self.inverse_options
        _, kwargs = self.add_argument_args

        if self.inverse_help:
            inverse_help = self.inverse_help
        elif self.help:
            help_ = kwargs['help']
            if help_.startswith('Don\'t '):
                inverse_help = help_[6:].capitalize()
            elif help_.startswith('Do not '):
                inverse_help = help_[7:].capitalize()
            else:
                first_letter = help_[0].lower()
                rest = help_[1:]
                inverse_help = 'Don\'t {first_letter}{rest}'.format_map(locals())
        else:
            inverse_help = self.help

        kwargs = kwargs.copy()
        kwargs['action'] = 'store_false'
        kwargs['help'] = inverse_help

        if self.is_bool_or:
            kwargs.pop('metavar')
            kwargs.pop('nargs')
            kwargs.pop('type')

        return args, kwargs

    def __str__(self):
        string = '{kind} arg: {self.name}{default}: type={type}'
        kind = 'Positional' if self.is_positional else 'Optional'
        has_default = self.default not in (EMPTY, None)
        default = '[={self.default}]'.format_map(locals()) if has_default else ''
        if self.is_bool:
            type = 'flag'
        elif self.is_bool_or:
            type = 'flag|{self.type.__name__}'.format_map(locals())
        elif self.type is None:
            type = None
        else:
            type = self.type.__name__
        return string.format_map(locals())


class HelpArg(Arg):

    def __init__(self, *, command):
        parameter = Parameter('help', POSITIONAL_OR_KEYWORD, default=False)
        super().__init__(
            command=command,
            parameter=parameter,
            name='help',
            container=None,
            type=bool,
            positional=None,
            default=False,
            choices=None,
            help=None,
            inverse_help=None,
            short_option='-h',
            long_option='--help',
            no_inverse=True,
            inverse_short_option=None,
            inverse_long_option=None,
            action=None,
            nargs=None,
            mutual_exclusion_group=None,
        )


class bool_or:

    """Used to indicate that an arg can be a flag or an option.

    Use like this::

        @command
        def local(config, cmd, hide: {'type': bool_or(str)} = False):
            "Run the specified command, possibly hiding its output."

    Allows for this::

        run local --hide all     # Hide everything
        run local --hide         # Hide everything with less effort
        run local --hide stdout  # Hide stdout only
        run local --no-hide      # Don't hide anything

    """

    type = None

    def __new__(cls, type, *, _type_cache={}):
        if type not in _type_cache:
            name = 'BoolOr{name}'.format(name=type.__name__.title())
            _type_cache[type] = builtins.type(name, (cls,), {'type': type})
        return _type_cache[type]


class BoolOrAction(argparse.Action):

    def __call__(self, parser, namespace, value, option_string=None):
        if value is None:
            value = True
        setattr(namespace, self.dest, value)


class ContainerAction(argparse.Action):

    @classmethod
    def from_container(cls, container_type):
        return type('ContainerAction', (cls,), {'container_type': container_type})

    def __call__(self, parser, namespace, values, option_string=None):
        items_to_add = []
        items = getattr(namespace, self.dest, self.container_type())
        if isinstance(items, Mapping):
            if items is not None:
                items_to_add.extend(items.items())
            for value in values:
                try:
                    name, value = value.split(':', 1)
                except ValueError:
                    message = (
                        'Bad format for {self.option_strings[0]}; '
                        'expected name:<value> but got: {value}')
                    raise CommandError(message.format_map(locals()))
                value = self.type(value)
                items_to_add.append((name, value))
            items = self.container_type(items_to_add)
        else:  # Sequence
            if items is not None:
                items_to_add.extend(items)
            items_to_add.extend(self.type(value) for value in values)
            items = self.container_type(items_to_add)
        setattr(namespace, self.dest, items)


def json_value(string):
    """Convert string to JSON if possible; otherwise, return as is."""
    try:
        string = json.loads(string)
    except ValueError:
        pass
    return string



import os
from subprocess import CompletedProcess

from .exc import RunCommandsError
from .util import cached_property


class Result(RunCommandsError):

    def __init__(self, args, return_code, stdout, stderr):
        self.args = args
        self.args_str = args if isinstance(args, str) else ' '.join(args)
        self.return_code = return_code
        self.stdout = stdout
        self.stderr = stderr
        self.succeeded = self.return_code == 0
        self.failed = not self.succeeded

    @classmethod
    def from_subprocess_result(cls, result: CompletedProcess):
        return cls(
            result.args,
            result.returncode,
            result.stdout,
            result.stderr,
        )

    @cached_property
    def stdout_lines(self):
        return self.stdout.splitlines() if self.stdout else []

    @cached_property
    def stderr_lines(self):
        return self.stderr.splitlines() if self.stderr else []

    def __bool__(self):
        return self.succeeded

    def __str__(self):
        output = (self.stderr if self.return_code else self.stdout) or '[NO OUTPUT]'
        if output.endswith(os.linesep):
            output = output[:-1]
        status = 'SUCCEEDED' if self.succeeded else 'FAILED'
        string = '{status} ({self.return_code}): {self.args_str} -> {output}'
        string = string.format_map(locals())
        return string

    def __repr__(self):
        return repr(str(self))



import os
from itertools import chain

from .exc import RunnerError
from .util import printer


class CommandRunner:

    """A runner for a given collection of commands."""

    def __init__(self, collection, debug=False):
        self.collection = collection
        self.debug = debug

    def run(self, argv):
        commands_to_run = self.get_commands_to_run(self.collection, argv)

        show_help_for = [c for c in commands_to_run if c.help_requested]
        if show_help_for:
            count = len(show_help_for)
            for command in show_help_for:
                if count > 1:
                    printer.hr('Help for', command.name, end=os.linesep * 2)
                command.show_help()
            return ()

        return tuple(command.run() for command in commands_to_run)

    def get_commands_to_run(self, collection, argv):
        debug = self.debug
        partition_args = self.partition_args
        commands_to_run = []
        while argv:
            command, command_argv = partition_args(collection, argv)
            command_to_run = CommandToRun(command, command_argv)
            commands_to_run.append(command_to_run)
            num_consumed = len(command_argv) + 1
            argv = argv[num_consumed:]
            if debug:
                printer.debug('Command to run:', command_to_run)
        return commands_to_run

    def partition_args(self, collection, args):
        name = args[0]

        try:
            command = collection[name]
        except KeyError:
            raise RunnerError('Unknown command: {name}'.format(name=name))

        args = args[1:]
        command_args = []
        partition = [command, command_args]

        prev_args = chain([None], args[:-1])

        for prev_arg, arg in zip(prev_args, args):
            if arg in collection:
                # Found a command name. If the previous arg wasn't an
                # option expecting a value, assume the command name is
                # the next command to run and not an option value or
                # positional.
                option = command.option_map.get(prev_arg)
                if option is None or not option.takes_value:
                    break
            if arg and arg[0] == ':' and arg != ':' and arg[1:] in collection:
                # Found an escaped command name. Unescape it.
                arg = arg[1:]
            command_args.append(arg)

        return partition

    def print_usage(self):
        if not self.collection:
            printer.warning('No commands available')
            return
        print('\nAvailable commands:\n')
        for name in sorted(self.collection):
            print('    {name}'.format(name=name))
        print('\nFor detailed help on a command: runcommands <command> --help')


class CommandToRun:

    __slots__ = ('name', 'command', 'argv', 'help_requested')

    def __init__(self, command, argv):
        argv = command.expand_short_options(argv)
        self.name = command.name
        self.command = command
        self.argv = argv

        # Ignore args after -- when determining whether help was
        # requested for a command because such args aren't command
        # options.
        try:
            dash_dash_index = argv.index('--')
        except ValueError:
            help_requested_argv = argv
        else:
            help_requested_argv = argv[:dash_dash_index]

        self.help_requested = '-h' in help_requested_argv or '--help' in help_requested_argv

    def run(self):
        argv_dict = self.command.parse_args(self.argv, False)
        return self.command.run(argv_dict)

    def show_help(self):
        print(self.command.help)

    def __repr__(self):
        return 'Command(name={self.name}, argv={self.argv})'.format(self=self)



from .misc import abort
from .printer import printer


def confirm(prompt='Really?', color='warning', yes_values=('y', 'yes'),
            abort_on_unconfirmed=False, abort_options=None):
    """Prompt for confirmation.

    Confirmation can be aborted by typing in a no value instead of one
    of the yes values or with Ctrl-C.

    Args:
        prompt (str): Prompt to present user ["Really?"]
        color (string|Color|bool) Color to print prompt string; can be
            ``False`` or ``None`` to print without color ["yellow"]
        yes_values (list[str]): Values user must type in to confirm
            [("y", "yes")]
        abort_on_unconfirmed (bool|int|str): When user does *not*
            confirm:

            - If this is an integer, print "Aborted" to stdout if
              it's 0 or to stderr if it's not 0 and then exit with
              this code
            - If this is a string, print it to stdout and exit with
              code 0
            - If this is ``True`` (or any other truthy value), print
              "Aborted" to stdout and exit with code 0

        abort_options (dict): Options to pass to :func:`abort` when not
            confirmed (these options will override any options set via
            ``abort_on_unconfirmed``)

    """
    if isinstance(yes_values, str):
        yes_values = (yes_values,)

    prompt = '{prompt} [{yes_value}/N] '.format(prompt=prompt, yes_value=yes_values[0])

    if color:
        prompt = printer.colorize(prompt, color=color)

    try:
        answer = input(prompt)
    except KeyboardInterrupt:
        print()
        confirmed = False
    else:
        answer = answer.strip().lower()
        confirmed = answer in yes_values

    # NOTE: The abort-on-unconfirmed logic is somewhat convoluted
    #       because of the special case for return code 0.

    do_abort_on_unconfirmed = not confirmed and (
        # True, non-zero return code, non-empty string, or any other
        # truthy value (in the manner of typical Python duck-typing)
        bool(abort_on_unconfirmed) or

        # Zero return code (special case)
        (abort_on_unconfirmed == 0 and abort_on_unconfirmed is not False)
    )

    if do_abort_on_unconfirmed:
        if abort_options is None:
            abort_options = {}

        if abort_on_unconfirmed is True:
            abort_options.setdefault('return_code', 0)
        elif isinstance(abort_on_unconfirmed, int):
            abort_options.setdefault('return_code', abort_on_unconfirmed)
        elif isinstance(abort_on_unconfirmed, str):
            abort_options.setdefault('message', abort_on_unconfirmed)
        else:
            abort_options.setdefault('return_code', 0)

        abort(**abort_options)

    return confirmed


def prompt(message, default=None, color=True):
    message = message.rstrip()
    if default is not None:
        default = default.rstrip()
        message = '%s [%s]' % (message, default)
    message = '%s ' % message
    if color is True:
        color = 'warning'
    if color:
        message = printer.colorize(message, color=color)
    try:
        value = input(message)
    except KeyboardInterrupt:
        print()
        abort()
    value = value.strip()
    if not value and default is not None:
        return default
    return value



import enum
import shutil
import sys
from typing import Mapping

from .enums import Color
from .misc import isatty


class ColorMap:

    def __init__(self, color_map: Mapping[str, str]):
        self.add_colors(color_map)

    def __getitem__(self, name: str):
        return getattr(self, name)

    def __setitem__(self, name: str, color: str):
        setattr(self, name, color)

    def add_colors(self, color_map: Mapping[str, str]):
        for name, color in color_map.items():
            setattr(self, name, color)


class Printer:

    # Symbolic name => color
    color_map = {
        'header': Color.magenta,
        'info': Color.blue,
        'success': Color.green,
        'echo': Color.cyan,
        'warning': Color.yellow,
        'error': Color.red,
        'danger': Color.red,
        'debug': Color.cyan,
    }

    def __init__(self, colors: enum.Enum = Color, color_map: Mapping = None, default_color=None):
        self.colors = colors
        self.color_map = ColorMap({color.name: color for color in colors})
        self.color_map.add_colors(self.__class__.color_map)
        if color_map:
            self.color_map.add_colors(color_map)
        self.default_color = self.get_color(default_color)

    def __call__(self, *args, **kwargs):
        self.print(*args, **kwargs)

    def get_color(self, color):
        if color is None:
            return self.color_map.none
        if isinstance(color, self.colors):
            return color
        try:
            return self.color_map[color]
        except KeyError:
            raise ValueError('Unknown color: {color}'.format(color=color)) from None

    def colorize(self, *args, color=None, sep=' ', end='reset'):
        color = self.get_color(color)
        args = (color,) + args
        string = []
        for arg in args[:-1]:
            string.append(str(arg))
            if not isinstance(arg, self.colors):
                string.append(sep)
        string.append(str(args[-1]))
        string = ''.join(string)
        if end:
            end = self.get_color(end)
            string = '{string}{end}'.format_map(locals())
        return string

    def print(self, *args, color=None, file=sys.stdout, **kwargs):
        if color is None:
            color = self.default_color
        if isatty(file):
            colorize_kwargs = kwargs.copy()
            colorize_kwargs.pop('end', None)
            colorize_kwargs.pop('flush', None)
            string = self.colorize(*args, color=color, **colorize_kwargs)
            print(string, file=file, **kwargs)
        else:
            args = [a for a in args if not isinstance(a, self.colors)]
            print(*args, file=file, **kwargs)

    def header(self, *args, color=None, **kwargs):
        if color is None:
            color = self.color_map.header
        self.print(*args, color=color, **kwargs)

    def info(self, *args, color=None, **kwargs):
        if color is None:
            color = self.color_map.info
        self.print(*args, color=color, **kwargs)

    def success(self, *args, color=None, **kwargs):
        if color is None:
            color = self.color_map.success
        self.print(*args, color=color, **kwargs)

    def echo(self, *args, color=None, **kwargs):
        if color is None:
            color = self.color_map.echo
        self.print(*args, color=color, **kwargs)

    def warning(self, *args, color=None, file=sys.stderr, **kwargs):
        if color is None:
            color = self.color_map.warning
        self.print(*args, color=color, file=file, **kwargs)

    def error(self, *args, color=None, file=sys.stderr, **kwargs):
        if color is None:
            color = self.color_map.error
        self.print(*args, color=color, file=file, **kwargs)

    def danger(self, *args, color=None, file=sys.stderr, **kwargs):
        if color is None:
            color = self.color_map.danger
        self.print(*args, color=color, file=file, **kwargs)

    def debug(self, *args, color=None, file=sys.stderr, **kwargs):
        if color is None:
            color = self.color_map.debug
        self.print(*args, color=color, file=file, **kwargs)

    def hr(self, *args, color=None, fill_char='=', **kwargs):
        if color is None:
            color = self.color_map.header
        hr = get_hr(fill_char)
        if args:
            sep = kwargs.get('sep') or ' '
            hr = hr[len(sep.join(args)) + len(sep):]
            prefix_kwargs = kwargs.copy()
            prefix_kwargs['sep'] = sep
            prefix_kwargs['end'] = sep
            self.print(*args, color=color, **prefix_kwargs)
        self.print(hr, color=color, **kwargs)


printer = Printer()


def get_hr(fill_char='='):
    term_size = shutil.get_terminal_size((80, 25))
    hr = fill_char * term_size.columns
    return hr



import importlib
import os

from .printer import printer


def abs_path(path, format_kwargs={}, relative_to=None, keep_slash=False):
    """Get abs. path for ``path``.

    ``path`` may be a relative or absolute file system path or an asset
    path. If ``path`` is already an abs. path, it will be returned as
    is. Otherwise, it will be converted into a normalized abs. path.

    If ``relative_to`` is passed *and* ``path`` is not absolute, the
    path will be joined to the specified prefix before it's made
    absolute.

    If ``path`` ends with a slash, it will be stripped unless
    ``keep_slash`` is set (for use with ``rsync``, for example).

    >>> file_path = os.path.normpath(__file__)
    >>> dir_name = os.path.dirname(file_path)
    >>> file_name = os.path.basename(file_path)
    >>> os.chdir(dir_name)
    >>>
    >>> abs_path(file_name) == file_path
    True
    >>> abs_path('runcommands.util:') == dir_name
    True
    >>> abs_path('runcommands.util:path.py') == file_path
    True
    >>> abs_path('/{xyz}', format_kwargs={'xyz': 'abc'})
    '/abc'
    >>> abs_path('banana', relative_to='/usr')
    '/usr/banana'
    >>> abs_path('/usr/banana/')
    '/usr/banana'
    >>> abs_path('banana/', relative_to='/usr', keep_slash=True)
    '/usr/banana/'
    >>> abs_path('runcommands.util:banana/', keep_slash=True) == (dir_name + '/banana/')
    True

    """
    if format_kwargs:
        path = path.format_map(format_kwargs)

    has_slash = path.endswith(os.sep)

    if os.path.isabs(path):
        path = os.path.normpath(path)
    elif ':' in path:
        path = asset_path(path, keep_slash=False)
    else:
        path = os.path.expanduser(path)
        if relative_to:
            path = os.path.join(relative_to, path)
        path = os.path.abspath(path)
        path = os.path.normpath(path)

    if has_slash and keep_slash:
        path = '{path}{slash}'.format(path=path, slash=os.sep)

    return path


def asset_path(path, format_kwargs={}, keep_slash=False):
    """Get absolute path to asset in package.

    ``path`` can be just a package name like 'package' or it can be
    a package name and a relative file system path like 'package:util'.

    If ``path`` ends with a slash, it will be stripped unless
    ``keep_slash`` is set (for use with ``rsync``, for example).

    >>> file_path = os.path.normpath(__file__)
    >>> dir_name = os.path.dirname(file_path)
    >>> file_name = os.path.basename(file_path)
    >>> os.chdir(dir_name)
    >>>
    >>> asset_path('runcommands.util') == dir_name
    True
    >>> asset_path('runcommands.util:path.py') == file_path
    True
    >>> asset_path('runcommands.util:{name}.py', format_kwargs={'name': 'path'}) == file_path
    True
    >>> asset_path('runcommands.util:dir/') == (dir_name + '/dir')
    True
    >>> asset_path('runcommands.util:dir/', keep_slash=True) == (dir_name + '/dir/')
    True

    """
    if format_kwargs:
        path = path.format_map(format_kwargs)

    has_slash = path.endswith(os.sep)

    if ':' in path:
        package_name, *rel_path = path.split(':', 1)
    else:
        package_name, rel_path = path, ()

    try:
        package = importlib.import_module(package_name)
    except ImportError:
        raise ValueError(
            'Could not get asset path for {path}; could not import package: {package_name}'
            .format_map(locals()))

    if not hasattr(package, '__file__'):
        raise ValueError("Can't compute path relative to namespace package")

    package_path = os.path.dirname(package.__file__)
    path = os.path.join(package_path, *rel_path)
    path = os.path.normpath(path)

    if has_slash and keep_slash:
        path = '{path}{slash}'.format(path=path, slash=os.sep)

    return path


def paths_to_str(paths, format_kwargs={}, delimiter=os.pathsep, asset_paths=False,
                 check_paths=False):
    """Convert ``paths`` to a single string.

    Args:
        paths (str|list): A string like "/a/path:/another/path" or
            a list of paths; may include absolute paths and/or asset
            paths; paths that are relative will be left relative
        format_kwargs (dict): Will be injected into each path
        delimiter (str): The string used to separate paths
        asset_paths (bool): Whether paths that look like asset paths
            will be converted to absolute paths
        check_paths (bool): Whether paths should be checked to ensure
            they exist

    """
    if not paths:
        return ''
    if isinstance(paths, str):
        paths = paths.split(delimiter)
    processed_paths = []
    for path in paths:
        original = path
        path = path.format_map(format_kwargs)
        if not os.path.isabs(path):
            if asset_paths and ':' in path:
                try:
                    path = asset_path(path)
                except ValueError:
                    path = None
        if path is not None and os.path.isdir(path):
            processed_paths.append(path)
        elif check_paths:
            f = locals()
            printer.warning('Path does not exist: {path} (from {original})'.format_map(f))
    return delimiter.join(processed_paths)
